/*
 * There used to be a note that said this:
 *
 * The author - Don Woods - apologises for the style of the code; it
 * is a result of running the original Fortran IV source through a
 * home-brew Fortran-to-C converter.
 *
 * Now that the code has been restructured into something much closer
 * to idiomatic C, the following is more appropriate:
 *
 * ESR apologizes for the remaing gotos (now confined to one function
 * in this file - there used to be over 350 of them, *everywhere*),
 * and for the offensive globals.  Applying the Structured Program
 * Theorem can be hard.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <getopt.h>
#include <signal.h>
#include <string.h>
#include "advent.h"
#include "dungeon.h"

#define DIM(a) (sizeof(a)/sizeof(a[0]))

// LCOV_EXCL_START
// exclude from coverage analysis because it requires interactivity to test
static void sig_handler(int signo)
{
    if (signo == SIGINT) {
        if (settings.logfp != NULL)
            fflush(settings.logfp);
    }
    exit(EXIT_FAILURE);
}
// LCOV_EXCL_STOP

/*
 * MAIN PROGRAM
 *
 *  Adventure (rev 2: 20 treasures)
 *  History: Original idea & 5-treasure version (adventures) by Willie Crowther
 *           15-treasure version (adventure) by Don Woods, April-June 1977
 *           20-treasure version (rev 2) by Don Woods, August 1978
 *		Errata fixed: 78/12/25
 *	     Revived 2017 as Open Adventure.
 */

static bool do_command(void);

int main(int argc, char *argv[])
{
    int ch;

    /*  Options. */

#ifndef ADVENT_NOSAVE
    const char* opts = "l:or:";
    const char* usage = "Usage: %s [-l logfilename] [-o] [-r restorefilename]\n";
    FILE *rfp = NULL;
#else
    const char* opts = "l:o";
    const char* usage = "Usage: %s [-l logfilename] [-o]\n";
#endif
    while ((ch = getopt(argc, argv, opts)) != EOF) {
        switch (ch) {
        case 'l':
            settings.logfp = fopen(optarg, "w");
            if (settings.logfp == NULL)
                fprintf(stderr,
                        "advent: can't open logfile %s for write\n",
                        optarg);
            signal(SIGINT, sig_handler);
            break;
        case 'o':
            settings.oldstyle = true;
            settings.prompt = false;
            break;
#ifndef ADVENT_NOSAVE
        case 'r':
            rfp = fopen(optarg, "r");
            if (rfp == NULL)
                fprintf(stderr,
                        "advent: can't open save file %s for read\n",
                        optarg);
            signal(SIGINT, sig_handler);
            break;
#endif
        default:
            fprintf(stderr,
                    usage, argv[0]);
            fprintf(stderr,
                    "        -l create a log file of your game named as specified'\n");
            fprintf(stderr,
                    "        -o 'oldstyle' (no prompt, no command editing, displays 'Initialising...')\n");
#ifndef ADVENT_NOSAVE
            fprintf(stderr,
                    "        -r restore from specified saved game file\n");
#endif
            exit(EXIT_FAILURE);
            break;
        }
    }

    /*  Initialize game variables */
    long seedval = initialise();

#ifndef ADVENT_NOSAVE
    if (!rfp) {
        game.novice = yes(arbitrary_messages[WELCOME_YOU], arbitrary_messages[CAVE_NEARBY], arbitrary_messages[NO_MESSAGE]);
        if (game.novice)
            game.limit = NOVICELIMIT;
    } else {
        restore(rfp);
    }
#endif

    if (settings.logfp)
        fprintf(settings.logfp, "seed %ld\n", seedval);

    /* interpret commands until EOF or interrupt */
    for (;;) {
        if (!do_command())
            break;
    }
    /* show score and exit */
    terminate(quitgame);
}

static bool fallback_handler(char *buf)
/* fallback handler for commands not handled by FORTRANish parser */
{
    long sv;
    if (sscanf(buf, "seed %ld", &sv) == 1) {
        set_seed(sv);
        printf("Seed set to %ld\n", sv);
        // autogenerated, so don't charge user time for it.
        --game.turns;
        return true;
    }
    return false;
}

/*  Check if this loc is eligible for any hints.  If been here long
 *  enough, display.  Ignore "HINTS" < 4 (special stuff, see database
 *  notes). */
static void checkhints(void)
{
    if (conditions[game.loc] >= game.conds) {
        for (int hint = 0; hint < NHINTS; hint++) {
            if (game.hinted[hint])
                continue;
            if (!CNDBIT(game.loc, hint + 1 + COND_HBASE))
                game.hintlc[hint] = -1;
            ++game.hintlc[hint];
            /*  Come here if he's been long enough at required loc(s) for some
             *  unused hint. */
            if (game.hintlc[hint] >= hints[hint].turns) {
                int i;

                switch (hint) {
                case 0:
                    /* cave */
                    if (game.prop[GRATE] == GRATE_CLOSED && !HERE(KEYS))
                        break;
                    game.hintlc[hint] = 0;
                    return;
                case 1:	/* bird */
                    if (game.place[BIRD] == game.loc && TOTING(ROD) && game.oldobj == BIRD)
                        break;
                    return;
                case 2:	/* snake */
                    if (HERE(SNAKE) && !HERE(BIRD))
                        break;
                    game.hintlc[hint] = 0;
                    return;
                case 3:	/* maze */
                    if (game.atloc[game.loc] == 0 &&
                        game.atloc[game.oldloc] == 0 &&
                        game.atloc[game.oldlc2] == 0 &&
                        game.holdng > 1)
                        break;
                    game.hintlc[hint] = 0;
                    return;
                case 4:	/* dark */
                    if (game.prop[EMERALD] != STATE_NOTFOUND && game.prop[PYRAMID] == STATE_NOTFOUND)
                        break;
                    game.hintlc[hint] = 0;
                    return;
                case 5:	/* witt */
                    break;
                case 6:	/* urn */
                    if (game.dflag == 0)
                        break;
                    game.hintlc[hint] = 0;
                    return;
                case 7:	/* woods */
                    if (game.atloc[game.loc] == 0 &&
                        game.atloc[game.oldloc] == 0 &&
                        game.atloc[game.oldlc2] == 0)
                        break;
                    return;
                case 8:	/* ogre */
                    i = atdwrf(game.loc);
                    if (i < 0) {
                        game.hintlc[hint] = 0;
                        return;
                    }
                    if (HERE(OGRE) && i == 0)
                        break;
                    return;
                case 9:	/* jade */
                    if (game.tally == 1 && game.prop[JADE] < 0)
                        break;
                    game.hintlc[hint] = 0;
                    return;
                default:
                    BUG(HINT_NUMBER_EXCEEDS_GOTO_LIST); // LCOV_EXCL_LINE
                    break;
                }

                /* Fall through to hint display */
                game.hintlc[hint] = 0;
                if (!yes(hints[hint].question, arbitrary_messages[NO_MESSAGE], arbitrary_messages[OK_MAN]))
                    return;
                rspeak(HINT_COST, hints[hint].penalty, hints[hint].penalty);
                game.hinted[hint] = yes(arbitrary_messages[WANT_HINT], hints[hint].hint, arbitrary_messages[OK_MAN]);
                if (game.hinted[hint] && game.limit > WARNTIME)
                    game.limit += WARNTIME * hints[hint].penalty;
            }
        }
    }
}

static bool spotted_by_pirate(int i)
{
    if (i != PIRATE)
        return false;

    /*  The pirate's spotted him.  He leaves him alone once we've
     *  found chest.  K counts if a treasure is here.  If not, and
     *  tally=1 for an unseen chest, let the pirate be spotted.  Note
     *  that game.place[CHEST] = LOC_NOWHERE might mean that he's thrown
     *  it to the troll, but in that case he's seen the chest
     *  (game.prop=0). */
    if (game.loc == game.chloc || game.prop[CHEST] >= 0)
        return true;
    int snarfed = 0;
    bool movechest = false, robplayer = false;
    for (int treasure = 1; treasure <= NOBJECTS; treasure++) {
        if (!objects[treasure].is_treasure)
            continue;
        /*  Pirate won't take pyramid from plover room or dark
         *  room (too easy!). */
        if (treasure == PYRAMID && (game.loc == objects[PYRAMID].plac || game.loc == objects[EMERALD].plac)) {
            continue;
        }
        if (TOTING(treasure) || HERE(treasure))
            ++snarfed;
        if (TOTING(treasure)) {
            movechest = true;
            robplayer = true;
        }
    }
    /* Force chest placement before player finds last treasure */
    if (game.tally == 1 && snarfed == 0 && game.place[CHEST] == LOC_NOWHERE && HERE(LAMP) && game.prop[LAMP] == LAMP_BRIGHT) {
        rspeak(PIRATE_SPOTTED);
        movechest = true;
    }
    /* Do things in this order (chest move before robbery) so chest is listed
     * last at the maze location. */
    if (movechest) {
        move(CHEST, game.chloc);
        move(MESSAG, game.chloc2);
        game.dloc[PIRATE] = game.chloc;
        game.odloc[PIRATE] = game.chloc;
        game.dseen[PIRATE] = false;
    } else {
        /* You might get a hint of the pirate's presence even if the
         * chest doesn't move... */
        if (game.odloc[PIRATE] != game.dloc[PIRATE] && PCT(20))
            rspeak(PIRATE_RUSTLES);
    }
    if (robplayer) {
        rspeak(PIRATE_POUNCES);
        for (int treasure = 1; treasure <= NOBJECTS; treasure++) {
            if (!objects[treasure].is_treasure)
                continue;
            if (!(treasure == PYRAMID && (game.loc == objects[PYRAMID].plac || game.loc == objects[EMERALD].plac))) {
                if (AT(treasure) && game.fixed[treasure] == 0)
                    carry(treasure, game.loc);
                if (TOTING(treasure))
                    drop(treasure, game.chloc);
            }
        }
    }

    return true;
}

static bool dwarfmove(void)
/* Dwarves move.  Return true if player survives, false if he dies. */
{
    int kk, stick, attack;
    long tk[21];

    /*  Dwarf stuff.  See earlier comments for description of
     *  variables.  Remember sixth dwarf is pirate and is thus
     *  very different except for motion rules. */

    /*  First off, don't let the dwarves follow him into a pit or
     *  a wall.  Activate the whole mess the first time he gets as
     *  far as the hall of mists (loc 15).  If game.newloc is
     *  forbidden to pirate (in particular, if it's beyond the
     *  troll bridge), bypass dwarf stuff.  That way pirate can't
     *  steal return toll, and dwarves can't meet the bear.  Also
     *  means dwarves won't follow him into dead end in maze, but
     *  c'est la vie.  They'll wait for him outside the dead
     *  end. */
    if (game.loc == 0 || FORCED(game.loc) || CNDBIT(game.newloc, COND_NOARRR))
        return true;

    /* Dwarf activity level ratchets up */
    if (game.dflag == 0) {
        if (INDEEP(game.loc))
            game.dflag = 1;
        return true;
    }

    /*  When we encounter the first dwarf, we kill 0, 1, or 2 of
     *  the 5 dwarves.  If any of the survivors is at loc,
     *  replace him with the alternate. */
    if (game.dflag == 1) {
        if (!INDEEP(game.loc) || (PCT(95) && (!CNDBIT(game.loc, COND_NOBACK) || PCT(85))))
            return true;
        game.dflag = 2;
        for (int i = 1; i <= 2; i++) {
            int j = 1 + randrange(NDWARVES - 1);
            if (PCT(50))
                game.dloc[j] = 0;
        }

        /* Alternate initial loc for dwarf, in case one of them
        *  starts out on top of the adventurer. */
        for (int i = 1; i <= NDWARVES - 1; i++) {
            if (game.dloc[i] == game.loc)
                game.dloc[i] = DALTLC; //
            game.odloc[i] = game.dloc[i];
        }
        rspeak(DWARF_RAN);
        drop(AXE, game.loc);
        return true;
    }

    /*  Things are in full swing.  Move each dwarf at random,
     *  except if he's seen us he sticks with us.  Dwarves stay
     *  deep inside.  If wandering at random, they don't back up
     *  unless there's no alternative.  If they don't have to
     *  move, they attack.  And, of course, dead dwarves don't do
     *  much of anything. */
    game.dtotal = 0;
    attack = 0;
    stick = 0;
    for (int i = 1; i <= NDWARVES; i++) {
        if (game.dloc[i] == 0)
            continue;
        /*  Fill tk array with all the places this dwarf might go. */
        unsigned int j = 1;
        kk = tkey[game.dloc[i]];
        if (kk != 0)
            do {
                game.newloc = travel[kk].dest;
                /* Have we avoided a dwarf encounter? */
                if (SPECIAL(game.newloc))
                    continue;
                else if (!INDEEP(game.newloc))
                    continue;
                else if (game.newloc == game.odloc[i])
                    continue;
                else if (j > 1 && game.newloc == tk[j - 1])
                    continue;
                else if (j >= DIM(tk) - 1)
                    continue;
                else if (game.newloc == game.dloc[i])
                    continue;
                else if (FORCED(game.newloc))
                    continue;
                else if (i == PIRATE && CNDBIT(game.newloc, COND_NOARRR))
                    continue;
                else if (travel[kk].nodwarves)
                    continue;
                tk[j++] = game.newloc;
            } while
            (!travel[kk++].stop);
        tk[j] = game.odloc[i];
        if (j >= 2)
            --j;
        j = 1 + randrange(j);
        game.odloc[i] = game.dloc[i];
        game.dloc[i] = tk[j];
        game.dseen[i] = (game.dseen[i] && INDEEP(game.loc)) || (game.dloc[i] == game.loc || game.odloc[i] == game.loc);
        if (!game.dseen[i])
            continue;
        game.dloc[i] = game.loc;
        if (spotted_by_pirate(i))
            continue;
        /* This threatening little dwarf is in the room with him! */
        ++game.dtotal;
        if (game.odloc[i] == game.dloc[i]) {
            ++attack;
            if (game.knfloc >= 0)
                game.knfloc = game.loc;
            if (randrange(1000) < 95 * (game.dflag - 2))
                ++stick;
        }
    }

    /*  Now we know what's happening.  Let's tell the poor sucker about it.
     *  Note that various of the "knife" messages must have specific relative
     *  positions in the rspeak database. */
    if (game.dtotal == 0)
        return true;
    rspeak(game.dtotal == 1 ? DWARF_SINGLE : DWARF_PACK, game.dtotal);
    if (attack == 0)
        return true;
    if (game.dflag == 2)
        game.dflag = 3;
    if (attack > 1) {
        rspeak(THROWN_KNIVES, attack);
        rspeak(stick > 1 ? MULTIPLE_HITS : (stick == 1 ? ONE_HIT : NONE_HIT), stick);
    } else {
        rspeak(KNIFE_THROWN);
        rspeak(MISSES_YOU);
    }
    if (stick == 0)
        return true;
    game.oldlc2 = game.loc;
    return false;
}

/*  "You're dead, Jim."
 *
 *  If the current loc is zero, it means the clown got himself killed.
 *  We'll allow this maxdie times.  NDEATHS is automatically set based
 *  on the number of snide messages available.  Each death results in
 *  a message (obituaries[n]) which offers reincarnation; if accepted,
 *  this results in message obituaries[0], obituaries[2], etc.  The
 *  last time, if he wants another chance, he gets a snide remark as
 *  we exit.  When reincarnated, all objects being carried get dropped
 *  at game.oldlc2 (presumably the last place prior to being killed)
 *  without change of props.  The loop runs backwards to assure that
 *  the bird is dropped before the cage.  (This kluge could be changed
 *  once we're sure all references to bird and cage are done by
 *  keywords.)  The lamp is a special case (it wouldn't do to leave it
 *  in the cave). It is turned off and left outside the building (only
 *  if he was carrying it, of course).  He himself is left inside the
 *  building (and heaven help him if he tries to xyzzy back into the
 *  cave without the lamp!).  game.oldloc is zapped so he can't just
 *  "retreat". */

static void croak(void)
/*  Okay, he's dead.  Let's get on with it. */
{
    const char* query = obituaries[game.numdie].query;
    const char* yes_response = obituaries[game.numdie].yes_response;
    ++game.numdie;
    if (game.closng) {
        /*  He died during closing time.  No resurrection.  Tally up a
         *  death and exit. */
        rspeak(DEATH_CLOSING);
        terminate(endgame);
    } else if (game.numdie == NDEATHS || !yes(query, yes_response, arbitrary_messages[OK_MAN]))
        terminate(endgame);
    else {
        game.place[WATER] = game.place[OIL] = LOC_NOWHERE;
        if (TOTING(LAMP))
            game.prop[LAMP] = LAMP_DARK;
        for (int j = 1; j <= NOBJECTS; j++) {
            int i = NOBJECTS + 1 - j;
            if (TOTING(i)) {
                /* Always leave lamp where it's accessible aboveground */
                drop(i, (i == LAMP) ? LOC_START : game.oldlc2);
            }
        }
        game.loc = LOC_BUILDING;
        game.oldloc = game.loc;
    }
}

static bool traveleq(long a, long b)
/* Are two travel entries equal for purposes of skip after failed condition? */
{
    return (travel[a].cond == travel[b].cond)
           && (travel[a].dest == travel[b].dest);
}

/*  Given the current location in "game.loc", and a motion verb number in
 *  "motion", put the new location in "game.newloc".  The current loc is saved
 *  in "game.oldloc" in case he wants to retreat.  The current
 *  game.oldloc is saved in game.oldlc2, in case he dies.  (if he
 *  does, game.newloc will be limbo, and game.oldloc will be what killed
 *  him, so we need game.oldlc2, which is the last place he was
 *  safe.) */

static bool playermove( int motion)
{
    int scratchloc, travel_entry = tkey[game.loc];
    game.newloc = game.loc;
    if (travel_entry == 0)
        BUG(LOCATION_HAS_NO_TRAVEL_ENTRIES); // LCOV_EXCL_LINE
    if (motion == NUL)
        return true;
    else if (motion == BACK) {
        /*  Handle "go back".  Look for verb which goes from game.loc to
         *  game.oldloc, or to game.oldlc2 If game.oldloc has forced-motion.
         *  te_tmp saves entry -> forced loc -> previous loc. */
        motion = game.oldloc;
        if (FORCED(motion))
            motion = game.oldlc2;
        game.oldlc2 = game.oldloc;
        game.oldloc = game.loc;
        int spk = 0;
        if (motion == game.loc)
            spk = FORGOT_PATH;
        if (CNDBIT(game.loc, COND_NOBACK))
            spk = TWIST_TURN;
        if (spk == 0) {
            int te_tmp = 0;
            for (;;) {
                scratchloc = travel[travel_entry].dest;
                if (scratchloc != motion) {
                    if (!SPECIAL(scratchloc)) {
                        if (FORCED(scratchloc) && travel[tkey[scratchloc]].dest == motion)
                            te_tmp = travel_entry;
                    }
                    if (!travel[travel_entry].stop) {
                        ++travel_entry;	/* go to next travel entry for this location */
                        continue;
                    }
                    /* we've reached the end of travel entries for game.loc */
                    travel_entry = te_tmp;
                    if (travel_entry == 0) {
                        rspeak(NOT_CONNECTED);
                        return true;
                    }
                }

                motion = travel[travel_entry].motion;
                travel_entry = tkey[game.loc];
                break; /* fall through to ordinary travel */
            }
        } else {
            rspeak(spk);
            return true;
        }
    } else if (motion == LOOK) {
        /*  Look.  Can't give more detail.  Pretend it wasn't dark
         *  (though it may now be dark) so he won't fall into a
         *  pit while staring into the gloom. */
        if (game.detail < 3)
            rspeak(NO_MORE_DETAIL);
        ++game.detail;
        game.wzdark = false;
        game.abbrev[game.loc] = 0;
        return true;
    } else if (motion == CAVE) {
        /*  Cave.  Different messages depending on whether above ground. */
        rspeak((OUTSID(game.loc) && game.loc != LOC_GRATE) ? FOLLOW_STREAM : NEED_DETAIL);
        return true;
    } else {
        /* none of the specials */
        game.oldlc2 = game.oldloc;
        game.oldloc = game.loc;
    }

    /* Look for a way to fulfil the motion verb passed in - travel_entry indexes
     * the beginning of the motion entries for here (game.loc). */
    for (;;) {
        if (T_TERMINATE(travel[travel_entry]) || travel[travel_entry].motion == motion)
            break;
        if (travel[travel_entry].stop) {
            /*  Couldn't find an entry matching the motion word passed
             *  in.  Various messages depending on word given. */
            int spk = CANT_APPLY;
            if (motion >= EAST && motion <= NW)
                spk = BAD_DIRECTION;
            if (motion == UP || motion == DOWN)
                spk = BAD_DIRECTION;
            if (motion == FORWARD || motion == LEFT || motion == RIGHT)
                spk = UNSURE_FACING;
            if (motion == OUTSIDE || motion == INSIDE)
                spk = NO_INOUT_HERE;
            if (motion == XYZZY || motion == PLUGH)
                spk = NOTHING_HAPPENS;
            if (motion == CRAWL)
                spk = WHICH_WAY;
            rspeak(spk);
            return true;
        }
        ++travel_entry;
    }

    /* (ESR) We've found a destination that goes with the motion verb.
     * Next we need to check any conditional(s) on this destination, and
     * possibly on following entries. */
    /* FIXME: Magic numbers related to move opcodes */
    do {
        for (;;) { /* L12 loop */
            for (;;) {
                long cond = travel[travel_entry].cond;
                long arg = MOD(cond, 100);
                if (!SPECIAL(cond)) {
                    /* YAML N and [pct N] conditionals */
                    if (cond <= 100) {
                        if (cond == 0 || PCT(cond))
                            break;
                        /* else fall through */
                    }
                    /* YAML [with OBJ] clause */
                    if (TOTING(arg) || (cond > 200 && AT(arg)))
                        break;
                    /* else fall through to check [not OBJ STATE] */
                } else if (game.prop[arg] != cond / 100 - 3)
                    break;

                /* We arrive here on conditional failure.
                 * Skip to next non-matching destination */
                long te_tmp = travel_entry;
                do {
                    if (travel[te_tmp].stop)
                        BUG(CONDITIONAL_TRAVEL_ENTRY_WITH_NO_ALTERATION); // LCOV_EXCL_LINE
                    ++te_tmp;
                } while
                (traveleq(travel_entry, te_tmp));
                travel_entry = te_tmp;
            }

            /* Found an eligible rule, now execute it */
            game.newloc = travel[travel_entry].dest;
            if (!SPECIAL(game.newloc))
                return true;

            if (game.newloc > 500) {
                /* Execute a speak rule */
                rspeak(L_SPEAK(game.newloc));
                game.newloc = game.loc;
                return true;
            } else {
                game.newloc -= SPECIALBASE;
                switch (game.newloc) {
                case 1:
                    /* Travel 301.  Plover-alcove passage.  Can carry only
                     * emerald.  Note: travel table must include "useless"
                     * entries going through passage, which can never be used
                     * for actual motion, but can be spotted by "go back". */
                    /* FIXME: Arithmetic on location numbers */
                    game.newloc = 99 + 100 - game.loc;
                    if (game.holdng > 1 || (game.holdng == 1 && !TOTING(EMERALD))) {
                        game.newloc = game.loc;
                        rspeak(MUST_DROP);
                    }
                    return true;
                case 2:
                    /* Travel 302.  Plover transport.  Drop the
                     * emerald (only use special travel if toting
                     * it), so he's forced to use the plover-passage
                     * to get it out.  Having dropped it, go back and
                     * pretend he wasn't carrying it after all. */
                    drop(EMERALD, game.loc);
                    int te_tmp = travel_entry;
                    do {
                        if (travel[te_tmp].stop)
                            BUG(CONDITIONAL_TRAVEL_ENTRY_WITH_NO_ALTERATION); // LCOV_EXCL_LINE
                        ++te_tmp;
                    } while
                    (traveleq(travel_entry, te_tmp));
                    travel_entry = te_tmp;
                    continue; /* goto L12 */
                case 3:
                    /* Travel 303.  Troll bridge.  Must be done only
                     * as special motion so that dwarves won't wander
                     * across and encounter the bear.  (They won't
                     * follow the player there because that region is
                     * forbidden to the pirate.)  If
                     * game.prop(TROLL)=1, he's crossed since paying,
                     * so step out and block him.  (standard travel
                     * entries check for game.prop(TROLL)=0.)  Special
                     * stuff for bear. */
                    if (game.prop[TROLL] == TROLL_PAIDONCE) {
		      pspeak(TROLL, look, TROLL_PAIDONCE, true);
                        game.prop[TROLL] = TROLL_UNPAID;
                        move(TROLL2, 0);
                        move(TROLL2 + NOBJECTS, 0);
                        move(TROLL, objects[TROLL].plac);
                        move(TROLL + NOBJECTS, objects[TROLL].fixd);
                        juggle(CHASM);
                        game.newloc = game.loc;
                        return true;
                    } else {
                        game.newloc = objects[TROLL].plac + objects[TROLL].fixd - game.loc;
                        if (game.prop[TROLL] == TROLL_UNPAID)
                            game.prop[TROLL] = TROLL_PAIDONCE;
                        if (!TOTING(BEAR))
                            return true;
                        rspeak(BRIDGE_COLLAPSE);
                        game.prop[CHASM] = BRIDGE_WRECKED;
                        game.prop[TROLL] = TROLL_GONE;
                        drop(BEAR, game.newloc);
                        game.fixed[BEAR] = -1;
                        game.prop[BEAR] = BEAR_DEAD;
                        game.oldlc2 = game.newloc;
                        croak();
                        return true;
                    }
                default:
                    BUG(SPECIAL_TRAVEL_500_GT_L_GT_300_EXCEEDS_GOTO_LIST); // LCOV_EXCL_LINE
                }
            }
            break; /* Leave L12 loop */
        }
    } while
    (false);
}

static bool closecheck(void)
/*  Handle the closing of the cave.  The cave closes "clock1" turns
 *  after the last treasure has been located (including the pirate's
 *  chest, which may of course never show up).  Note that the
 *  treasures need not have been taken yet, just located.  Hence
 *  clock1 must be large enough to get out of the cave (it only ticks
 *  while inside the cave).  When it hits zero, we branch to 10000 to
 *  start closing the cave, and then sit back and wait for him to try
 *  to get out.  If he doesn't within clock2 turns, we close the cave;
 *  if he does try, we assume he panics, and give him a few additional
 *  turns to get frantic before we close.  When clock2 hits zero, we
 *  transport him into the final puzzle.  Note that the puzzle depends
 *  upon all sorts of random things.  For instance, there must be no
 *  water or oil, since there are beanstalks which we don't want to be
 *  able to water, since the code can't handle it.  Also, we can have
 *  no keys, since there is a grate (having moved the fixed object!)
 *  there separating him from all the treasures.  Most of these
 *  problems arise from the use of negative prop numbers to suppress
 *  the object descriptions until he's actually moved the objects. */
{
    if (game.tally == 0 && INDEEP(game.loc) && game.loc != LOC_Y2)
        --game.clock1;

    /*  When the first warning comes, we lock the grate, destroy
     *  the bridge, kill all the dwarves (and the pirate), remove
     *  the troll and bear (unless dead), and set "closng" to
     *  true.  Leave the dragon; too much trouble to move it.
     *  from now until clock2 runs out, he cannot unlock the
     *  grate, move to any location outside the cave, or create
     *  the bridge.  Nor can he be resurrected if he dies.  Note
     *  that the snake is already gone, since he got to the
     *  treasure accessible only via the hall of the mountain
     *  king. Also, he's been in giant room (to get eggs), so we
     *  can refer to it.  Also also, he's gotten the pearl, so we
     *  know the bivalve is an oyster.  *And*, the dwarves must
     *  have been activated, since we've found chest. */
    if (game.clock1 == 0) {
        game.prop[GRATE] = GRATE_CLOSED;
        game.prop[FISSURE] = UNBRIDGED;
        for (int i = 1; i <= NDWARVES; i++) {
            game.dseen[i] = false;
            game.dloc[i] = 0;
        }
        move(TROLL, 0);
        move(TROLL + NOBJECTS, 0);
        move(TROLL2, objects[TROLL].plac);
        move(TROLL2 + NOBJECTS, objects[TROLL].fixd);
        juggle(CHASM);
        if (game.prop[BEAR] != BEAR_DEAD)
            DESTROY(BEAR);
        game.prop[CHAIN] = CHAIN_HEAP;
        game.fixed[CHAIN] = CHAIN_HEAP;
        game.prop[AXE] = 0;
        game.fixed[AXE] = 0;
        rspeak(CAVE_CLOSING);
        game.clock1 = -1;
        game.closng = true;
        return true;
    } else if (game.clock1 < 0)
        --game.clock2;
    if (game.clock2 == 0) {
        /*  Once he's panicked, and clock2 has run out, we come here
         *  to set up the storage room.  The room has two locs,
         *  hardwired as LOC_NE and LOC_SW.  At the ne end, we
         *  place empty bottles, a nursery of plants, a bed of
         *  oysters, a pile of lamps, rods with stars, sleeping
         *  dwarves, and him.  At the sw end we place grate over
         *  treasures, snake pit, covey of caged birds, more rods, and
         *  pillows.  A mirror stretches across one wall.  Many of the
         *  objects come from known locations and/or states (e.g. the
         *  snake is known to have been destroyed and needn't be
         *  carried away from its old "place"), making the various
         *  objects be handled differently.  We also drop all other
         *  objects he might be carrying (lest he have some which
         *  could cause trouble, such as the keys).  We describe the
         *  flash of light and trundle back. */
        game.prop[BOTTLE] = put(BOTTLE, LOC_NE, EMPTY_BOTTLE);
        game.prop[PLANT] = put(PLANT, LOC_NE, 0);
        game.prop[OYSTER] = put(OYSTER, LOC_NE, 0);
        game.prop[LAMP] = put(LAMP, LOC_NE, 0);
        game.prop[ROD] = put(ROD, LOC_NE, 0);
        game.prop[DWARF] = put(DWARF, LOC_NE, 0);
        game.loc = LOC_NE;
        game.oldloc = LOC_NE;
        game.newloc = LOC_NE;
        /*  Leave the grate with normal (non-negative) property.
         *  Reuse sign. */
        put(GRATE, LOC_SW, 0);
        put(SIGN, LOC_SW, 0);
        game.prop[SIGN] = ENDGAME_SIGN;
        game.prop[SNAKE] = put(SNAKE, LOC_SW, 1);
        game.prop[BIRD] = put(BIRD, LOC_SW, 1);
        game.prop[CAGE] = put(CAGE, LOC_SW, 0);
        game.prop[ROD2] = put(ROD2, LOC_SW, 0);
        game.prop[PILLOW] = put(PILLOW, LOC_SW, 0);

        game.prop[MIRROR] = put(MIRROR, LOC_NE, 0);
        game.fixed[MIRROR] = LOC_SW;

        for (int i = 1; i <= NOBJECTS; i++) {
            if (TOTING(i))
                DESTROY(i);
        }

        rspeak(CAVE_CLOSED);
        game.closed = true;
        return true;
    }

    return false;
}

static void lampcheck(void)
/* Check game limit and lamp timers */
{
    if (game.prop[LAMP] == LAMP_BRIGHT)
        --game.limit;

    /*  Another way we can force an end to things is by having the
     *  lamp give out.  When it gets close, we come here to warn him.
     *  First following arm checks if the lamp and fresh batteries are
     *  here, in which case we replace the batteries and continue.
     *  Second is for other cases of lamp dying.  Eve after it goes
     *  out, he can explore outside for a while if desired. */
    if (game.limit <= WARNTIME && HERE(BATTERY) && game.prop[BATTERY] == FRESH_BATTERIES && HERE(LAMP)) {
        rspeak(REPLACE_BATTERIES);
        game.prop[BATTERY] = DEAD_BATTERIES;
        if (TOTING(BATTERY))
            drop(BATTERY, game.loc);
        game.limit += BATTERYLIFE;
        game.lmwarn = false;
    } else if (game.limit == 0) {
        game.limit = -1;
        game.prop[LAMP] = LAMP_DARK;
        if (HERE(LAMP))
            rspeak(LAMP_OUT);
    } else if (game.limit <= WARNTIME) {
        if (!game.lmwarn && HERE(LAMP)) {
            game.lmwarn = true;
            int spk = GET_BATTERIES;
            if (game.place[BATTERY] == LOC_NOWHERE)
                spk = LAMP_DIM;
            if (game.prop[BATTERY] == DEAD_BATTERIES)
                spk = MISSING_BATTERIES;
            rspeak(spk);
        }
    }
}

static void listobjects(void)
/*  Print out descriptions of objects at this location.  If
 *  not closing and property value is negative, tally off
 *  another treasure.  Rug is special case; once seen, its
 *  game.prop is RUG_DRAGON (dragon on it) till dragon is killed.
 *  Similarly for chain; game.prop is initially CHAINING_BEAR (locked to
 *  bear).  These hacks are because game.prop=0 is needed to
 *  get full score. */
{
    if (!DARK(game.loc)) {
        ++game.abbrev[game.loc];
        for (int i = game.atloc[game.loc]; i != 0; i = game.link[i]) {
            long obj = i;
            if (obj > NOBJECTS)
                obj = obj - NOBJECTS;
            if (obj == STEPS && TOTING(NUGGET))
                continue;
            if (game.prop[obj] < 0) {
                if (game.closed)
                    continue;
                game.prop[obj] = 0;
                if (obj == RUG)
                    game.prop[RUG] = RUG_DRAGON;
                if (obj == CHAIN)
                    game.prop[CHAIN] = CHAINING_BEAR;
                --game.tally;
                /*  Note: There used to be a test here to see whether the
                 *  player had blown it so badly that he could never ever see
                 *  the remaining treasures, and if so the lamp was zapped to
                 *  35 turns.  But the tests were too simple-minded; things
                 *  like killing the bird before the snake was gone (can never
                 *  see jewelry), and doing it "right" was hopeless.  E.G.,
                 *  could cross troll bridge several times, using up all
                 *  available treasures, breaking vase, using coins to buy
                 *  batteries, etc., and eventually never be able to get
                 *  across again.  If bottle were left on far side, could then
                 *  never get eggs or trident, and the effects propagate.  So
                 *  the whole thing was flushed.  anyone who makes such a
                 *  gross blunder isn't likely to find everything else anyway
                 *  (so goes the rationalisation). */
            }
            int kk = game.prop[obj];
            if (obj == STEPS && game.loc == game.fixed[STEPS])
                kk = 1;
            pspeak(obj, look, kk, true);
        }
    }
}

static bool do_command()
/* Get and execute a command */
{
    long V1, V2;
    long kmod, defn;
    static long igo = 0;
    static struct command_t command;
    command.verb = 0;

    /*  Can't leave cave once it's closing (except by main office). */
    if (OUTSID(game.newloc) && game.newloc != 0 && game.closng) {
        rspeak(EXIT_CLOSED);
        game.newloc = game.loc;
        if (!game.panic)
            game.clock2 = PANICTIME;
        game.panic = true;
    }

    /*  See if a dwarf has seen him and has come from where he
     *  wants to go.  If so, the dwarf's blocking his way.  If
     *  coming from place forbidden to pirate (dwarves rooted in
     *  place) let him get out (and attacked). */
    if (game.newloc != game.loc && !FORCED(game.loc) && !CNDBIT(game.loc, COND_NOARRR)) {
        for (size_t i = 1; i <= NDWARVES - 1; i++) {
            if (game.odloc[i] == game.newloc && game.dseen[i]) {
                game.newloc = game.loc;
                rspeak(DWARF_BLOCK);
                break;
            }
        }
    }
    game.loc = game.newloc;

    if (!dwarfmove())
        croak();

    /*  Describe the current location and (maybe) get next command. */

    for (;;) {
        if (game.loc == 0)
            croak();
        const char* msg = locations[game.loc].description.small;
        if (MOD(game.abbrev[game.loc], game.abbnum) == 0 || msg == 0)
            msg = locations[game.loc].description.big;
        if (!FORCED(game.loc) && DARK(game.loc)) {
            /*  The easiest way to get killed is to fall into a pit in
             *  pitch darkness. */
            if (game.wzdark && PCT(35)) {
                rspeak(PIT_FALL);
                game.oldlc2 = game.loc;
                croak();
                continue;	/* back to top of main interpreter loop */
            }
            msg = arbitrary_messages[PITCH_DARK];
        }
        if (TOTING(BEAR))
            rspeak(TAME_BEAR);
        speak(msg);
        if (FORCED(game.loc)) {
            if (playermove(HERE))
                return true;
            else
                continue;	/* back to top of main interpreter loop */
        }
        if (game.loc == LOC_Y2 && PCT(25) && !game.closng)
            rspeak(SAYS_PLUGH);

        listobjects();

L2012:
        command.verb = 0;
        game.oldobj = command.obj;
        command.obj = 0;

L2600:
        checkhints();

        /*  If closing time, check for any objects being toted with
         *  game.prop < 0 and set the prop to -1-game.prop.  This way
         *  objects won't be described until they've been picked up
         *  and put down separate from their respective piles.  Don't
         *  tick game.clock1 unless well into cave (and not at Y2). */
        if (game.closed) {
            if (game.prop[OYSTER] < 0 && TOTING(OYSTER))
	      pspeak(OYSTER, look, 1, true);
            for (size_t i = 1; i <= NOBJECTS; i++) {
                if (TOTING(i) && game.prop[i] < 0)
                    game.prop[i] = -1 - game.prop[i];
            }
        }
        game.wzdark = DARK(game.loc);
        if (game.knfloc > 0 && game.knfloc != game.loc)
            game.knfloc = 0;

        /* This is where we get a new command from the user */
        char* input;
        char inputbuf[LINESIZE];

        for (;;) {
            input = get_input();
            if (input == NULL)
                return (false);
            if (word_count(input) > 2) {
                rspeak(TWO_WORDS);
                continue;
            }
            if (strcmp(input, "") != 0)
                break;
        }

        strncpy(inputbuf, input, LINESIZE - 1);
        free(input);

        long tokens[4];
        tokenize(inputbuf, tokens);
        command.wd1 = tokens[0];
        command.wd1x = tokens[1];
        command.wd2 = tokens[2];
        command.wd2x = tokens[3];

        /*  Every input, check "game.foobar" flag.  If zero, nothing's
         *  going on.  If pos, make neg.  If neg, he skipped a word,
         *  so make it zero. */
L2607:
        game.foobar = (game.foobar > 0 ? -game.foobar : 0);
        ++game.turns;

        /* If a turn threshold has been met, apply penalties and tell
         * the player about it. */
        for (int i = 0; i < NTHRESHOLDS; ++i) {
            if (game.turns == turn_thresholds[i].threshold + 1) {
                game.trnluz += turn_thresholds[i].point_loss;
                speak(turn_thresholds[i].message);
            }
        }

        if (command.verb == SAY && command.wd2 > 0)
            command.verb = 0;
        if (command.verb == SAY) {
            command.part = transitive;
            goto Laction;
        }
        if (closecheck()) {
            if (game.closed)
                return true;
        } else
            lampcheck();

        char word1[6];
        char word2[6];
        packed_to_token(command.wd1, word1);
        packed_to_token(command.wd2, word2);
        V1 = get_vocab_id(word1);
        V2 = get_vocab_id(word2);
        if (V1 == ENTER && (V2 == STREAM || V2 == PROMOTE_WORD(WATER))) {
            if (LIQLOC(game.loc) == WATER) {
                rspeak(FEET_WET);
            } else {
                rspeak(WHERE_QUERY);
            }
            goto L2012;
        }
        if (V1 == ENTER && command.wd2 > 0) {
            command.wd1 = command.wd2;
            command.wd1x = command.wd2x;
            wordclear(&command.wd2);
        } else {
            /* FIXME: Magic numbers related to vocabulary */
            if (!((V1 != PROMOTE_WORD(WATER) && V1 != PROMOTE_WORD(OIL)) ||
                  (V2 != PROMOTE_WORD(PLANT) && V2 != PROMOTE_WORD(DOOR)))) {
                if (AT(DEMOTE_WORD(V2)))
                    command.wd2 = token_to_packed("POUR");
            }
            if (V1 == PROMOTE_WORD(CAGE) && V2 == PROMOTE_WORD(BIRD) && HERE(CAGE) && HERE(BIRD))
                command.wd1 = token_to_packed("CATCH");
        }
L2620:
        if (wordeq(command.wd1, token_to_packed("WEST"))) {
            ++game.iwest;
            if (game.iwest == 10)
                rspeak(W_IS_WEST);
        }
        if (wordeq(command.wd1, token_to_packed("GO")) && !wordempty(command.wd2)) {
            if (++igo == 10)
                rspeak(GO_UNNEEDED);
        }
Lookup:
        packed_to_token(command.wd1, word1);
        defn = get_vocab_id(word1);
        if (defn == -1) {
            /* Gee, I don't understand. */
            if (fallback_handler(inputbuf))
                continue;
            rspeak(DONT_KNOW, command.wd1, command.wd1x);
            goto L2600;
        }
        /* FIXME: magic numbers related to vocabulary */
        kmod = MOD(defn, 1000);
        switch (defn / 1000) {
        case 0:
            if (playermove(kmod))
                return true;
            else
                continue;	/* back to top of main interpreter loop */
        case 1:
            command.part = unknown;
            command.obj = kmod;
            break;
        case 2:
            command.part = intransitive;
            command.verb = kmod;
            break;
        case 3:
            speak(specials[kmod].message);
            goto L2012;
        default:
            BUG(VOCABULARY_TYPE_N_OVER_1000_NOT_BETWEEN_0_AND_3); // LCOV_EXCL_LINE
        }

Laction:
        switch (action(&command)) {
        case GO_TERMINATE:
            return true;
        case GO_MOVE:
            playermove(NUL);
            return true;
        case GO_TOP:
            continue;	/* back to top of main interpreter loop */
        case GO_CLEAROBJ:
            goto L2012;
        case GO_CHECKHINT:
            goto L2600;
        case GO_CHECKFOO:
            goto L2607;
        case GO_LOOKUP:
            goto Lookup;
        case GO_WORD2:
            /* Get second word for analysis. */
            command.wd1 = command.wd2;
            command.wd1x = command.wd2x;
            wordclear(&command.wd2);
            goto L2620;
        case GO_UNKNOWN:
            /*  Random intransitive verbs come here.  Clear obj just in case
             *  (see attack()). */
            rspeak(DO_WHAT, command.wd1, command.wd1x);
            command.obj = 0;
            goto L2600;
        case GO_DWARFWAKE:
            /*  Oh dear, he's disturbed the dwarves. */
            rspeak(DWARVES_AWAKEN);
            terminate(endgame);
        default:
            BUG(ACTION_RETURNED_PHASE_CODE_BEYOND_END_OF_SWITCH); // LCOV_EXCL_LINE
        }
    }
}

/* end */
